from . import writer
from . import data_array

import functools
import numpy as np


class VTKFile:
    """Generic VTK file"""

    def __init__(self, filename, filetype, compression):
        self.filetype = filetype
        self.filename = filename
        self.writer = writer.Writer(filetype, compression)

        # Center piece
        self.piece = self.writer.registerPiece()

        # Registering data elements
        self.point_data = self.piece.register('PointData')
        self.cell_data = self.piece.register('CellData')

    def addPointData(self, data_array, vtk_format='binary'):
        self.point_data.registerDataArray(data_array, vtk_format)

    def addCellData(self, data_array):
        self.cell_data.registerDataArray(data_array)

    def write(self):
        self.writer.registerAppend()
        self.writer.write(self.filename)

    def __enter__(self):
        return self

    def __exit__(self, *args):
        self.write()


class ImageData(VTKFile):
    """VTK Image data (coordinates given by a range and constant spacing)"""

    def __init__(self, filename, ranges, points, compression=None):
        VTKFile.__init__(self, filename, 'ImageData', compression)

        # Computing spacings
        spacings = [(x[1] - x[0]) / (n - 1) for x, n in zip(ranges, points)]

        # Filling in missing coordinates
        for _ in range(len(points), 3):
            points.append(1)

        # Setting extents, spacing and origin
        extent = functools.reduce(
            lambda x, y: x + "0 {} ".format(y-1), points, "")
        spacings = functools.reduce(
            lambda x, y: x + "{} ".format(y), spacings, "")
        origins = functools.reduce(
            lambda x, y: x + "{} ".format(y[0]), ranges, "")

        self.writer.setDataNodeAttributes({
            'WholeExtent': extent,
            'Spacing': spacings,
            'Origin': origins
        })

        self.piece.setAttributes({
            "Extent": extent
        })


class RectilinearGrid(VTKFile):
    """VTK Rectilinear grid (coordinates are given by 3 seperate ranges)"""

    def __init__(self, filename, coordinates, offsets=None, compression=None):
        VTKFile.__init__(self, filename, 'RectilinearGrid', compression)

        # Checking that we actually have a list or tuple
        if type(coordinates).__name__ == 'ndarray':
            coordinates = [coordinates]

        self.coordinates = list(coordinates)

        # Filling in missing coordinates
        for _ in range(len(self.coordinates), 3):
            self.coordinates.append(np.array([0.]))

        # Setting data extent
        extent = []

        for coord in self.coordinates:
            if coord.ndim != 1:
                raise Exception(
                    'Coordinate array should have only one dimension'
                    + ' (has {})'.format(coord.ndim))
            extent.append(coord.size-1)

        if offsets is None:
            offsets = [0] * len(extent)
        elif len(offsets) == len(coordinates):
            offsets = list(offsets)
            offsets += [0] * (len(extent) - len(offsets))
        else:
            raise Exception(
                'Size of offsets should '
                'match domain dimension {}'.format(len(coordinates)))

        def fold_extent(acc, couple):
            offset, extent = couple

            if offset != 0:
                offset -= 1
            return acc + "{} {} ".format(offset, offset+extent)

        # Create extent string with offsets
        self.extent = functools.reduce(fold_extent, zip(offsets, extent), "")

        self.writer.setDataNodeAttributes({
            "WholeExtent": self.extent
        })

        self.piece.setAttributes({
            "Extent": self.extent
        })

        # Registering coordinates
        coordinate_component = self.piece.register('Coordinates')

        for coord, prefix in zip(self.coordinates, ('x', 'y', 'z')):
            array = data_array.DataArray(coord, [0], prefix + '_coordinates')
            coordinate_component.registerDataArray(array)


class StructuredGrid(VTKFile):
    """VTK Structured grid (coordinates given by a single array of points)"""

    def __init__(self, filename, points, shape, compression=None):
        VTKFile.__init__(self, filename, 'StructuredGrid', compression)

        if points.ndim != 2:
            raise 'Points should be a 2D array'

        # Completing the missing coordinates
        points_3d = np.zeros((points.shape[0], 3))
        for i in range(points.shape[1]):
            points_3d[:, i] = points[:, i]

        extent = [n - 1 for n in shape]
        for i in range(len(extent), 3):
            extent.append(0)

        extent = functools.reduce(
            lambda x, y: x + "0 {} ".format(y), extent, "")
        self.writer.setDataNodeAttributes({
            "WholeExtent": extent
        })

        self.piece.setAttributes({
            "Extent": extent
        })

        points_component = self.piece.register('Points')
        points_component.registerDataArray(
            data_array.DataArray(points_3d, [0], 'points'))



import functools


class DataArray:
    """Class holding information on ndarray"""

    def __init__(self, data, spatial_axes, name='', components_order='C'):
        """
        Data array constructor

        :param data: the numpy array containing the data (possibly a view)
        :param spatial_axes: a container of ints that indicate which axes of
        the array correspond to space dimensions (in order)
        :param name: the name of the data
        :param components_order: the order of the non-spatial axes of the array
        """
        self.data = data
        axes = list(range(data.ndim))
        spatial_axes = list(spatial_axes)

        if data.ndim < len(spatial_axes):
            raise Exception('Dimensions of data smaller than space dimensions')

        for ax in spatial_axes:
            axes.remove(ax)

        nb_components = functools.reduce(
            lambda x, y: x * data.shape[y], axes, 1)

        if components_order == 'C':
            axes.reverse()
        else:
            raise Exception('Unrecognized components order')

        axes += spatial_axes

        # Hopefully this is a view
        self.flat_data = self.data.transpose(*axes).reshape(-1, order='F')

        self.attributes = {
            "Name": name,
            "type": str(self.flat_data.dtype).capitalize(),
            "NumberOfComponents": str(nb_components)
        }

    def __str__(self):
        return self.attributes.__str__()



from . import writer
from . import data_array
from . import vtk_files

import functools

from os.path import splitext
from mpi4py import MPI

MASTER_RANK = 0


class PRectilinearGrid(vtk_files.RectilinearGrid):
    """Rectilinear grid parallel writer"""

    parent = vtk_files.RectilinearGrid

    def __init__(self, filename, coordinates, offset, compression=None):
        self.pfilename = filename
        self.comm = MPI.COMM_WORLD
        self.rank = self.comm.Get_rank()

        # Construct name of piece file
        fragmented_name = splitext(filename)
        self.piece_name_template = (fragmented_name[0] +
                                    '_rank{rank}.' + fragmented_name[1][2:])

        # Name of piece file associated to rank
        self.parent.__init__(
            self,
            self.piece_name_template.format(rank=self.rank),
            coordinates,
            offset,
            compression
        )

        # Gather local data extents
        extents = self.comm.gather(self.extent, root=MASTER_RANK)
        self.init_master(extents)

    def init_master(self, extents):
        if not extents:
            return

        self.pwriter = writer.Writer('P' + self.filetype)
        data_node = self.pwriter.data_node
        self.ppoint_data = self.pwriter.registerComponent('PPointData',
                                                          data_node)
        self.pcell_data = self.pwriter.registerComponent('PCellData',
                                                         data_node)

        for rank, extent in enumerate(extents):
            self.pwriter.registerPiece({
                'Source': self.piece_name_template.format(rank=rank),
                'Extent': extent
            })

        # Convert string extents to list of integers
        extents = list(zip(*map(lambda x: list(map(int, x.split())), extents)))
        mins = list(map(min, extents[0::2]))
        maxs = list(map(max, extents[1::2]))

        extent = functools.reduce(
            lambda x, y: x + "{} {} ".format(y[0], y[0]+y[1]),
            zip(mins, maxs),
            ""
        )

        self.pwriter.setDataNodeAttributes({
            'WholeExtent': extent
        })

        # Register coordinates
        pcoordinates = self.pwriter.registerComponent('PCoordinates',
                                                      data_node)

        for coord, prefix in zip(self.coordinates, ('x', 'y', 'z')):
            array = data_array.DataArray(coord, [0], prefix + '_coordinates')
            pcoordinates.registerPDataArray(array)

    def addPointData(self, array):
        self.parent.addPointData(self, array)
        if self.rank == MASTER_RANK:
            self.ppoint_data.registerPDataArray(array)

    def addCellData(self, array):
        self.parent.addCellData(self, array)
        if self.rank == MASTER_RANK:
            self.pcell_data.registerPDataArray(array)

    def write(self):
        self.parent.write(self)
        if self.rank == MASTER_RANK:
            self.pwriter.write(self.pfilename)



import xml.dom.minidom as dom
import io
import zlib

import numpy as np

from functools import reduce
from base64 import b64encode


def setAttributes(node, attributes):
    """Set attributes of a node"""
    for item in attributes.items():
        node.setAttribute(*item)


def encodeArray(array, level):
    """Encode array data and header in base64."""
    def compress(array):
        """Compress array with zlib. Returns header and compressed data."""
        raw_data = memoryview(array)  # avoid a copy
        data_size = raw_data.nbytes

        max_block_size = 2**15

        # Enough blocks to span whole data
        nblocks = data_size // max_block_size + 1
        last_block_size = data_size % max_block_size

        # Compress regular blocks
        compressed_data = [
            zlib.compress(raw_data[i*max_block_size:(i+1)*max_block_size],
                          level)
            for i in range(nblocks-1)
        ]

        # Compress last (smaller) block
        compressed_data.append(
            zlib.compress(raw_data[-last_block_size:], level)
        )

        # Header data (cf https://vtk.org/Wiki/VTK_XML_Formats#Compressed_Data)
        usize = max_block_size
        psize = last_block_size
        csize = [len(x) for x in compressed_data]
        header = np.array([nblocks, usize, psize] + csize, dtype=np.uint32)
        return header.tobytes(), b"".join(compressed_data)

    def raw(array):
        """Returns header and array data in bytes."""
        header = np.array([array.nbytes], dtype=np.uint32)
        return header.tobytes(), memoryview(array)

    if level is not None:
        data = compress(array)
    else:
        data = raw(array)
    return "".join(map(lambda x: b64encode(x).decode(), data))


class Component:
    """Generic component class capable of registering sub-components"""

    def __init__(self, name, parent_node, writer):
        self.writer = writer
        self.document = writer.document
        self.node = self.document.createElement(name)
        parent_node.appendChild(self.node)

    def setAttributes(self, attributes):
        setAttributes(self.node, attributes)

    def register(self, name, attributes={}):
        """Register a sub-component"""
        if type(attributes) != dict:
            raise Exception(
                'Cannot register attributes of type ' + str(type(attributes)))
        sub_component = Component(name, self.node, self.writer)
        setAttributes(sub_component.node, attributes)
        return sub_component

    def _addArrayNodeData(self, data_array, node, vtk_format):
        if vtk_format == 'ascii':
            data_as_str = reduce(
                lambda x, y: x + str(y) + ' ', data_array.flat_data, "")
        elif vtk_format == 'binary':
            data_as_str = encodeArray(data_array.flat_data,
                                      self.writer.compression)
        else:
            raise Exception('Unsupported VTK Format "{}"'.format(vtk_format))

        node.appendChild(self.document.createTextNode(data_as_str))

    def _registerArrayComponent(self, array, name, vtk_format):
        attributes = array.attributes
        attributes['format'] = vtk_format
        return self.register(name, attributes)

    def registerDataArray(self, data_array, vtk_format='binary'):
        """Register a DataArray object"""
        component = self._registerArrayComponent(data_array,
                                                 'DataArray',
                                                 vtk_format)
        self._addArrayNodeData(data_array, component.node, vtk_format)

    def registerPDataArray(self, data_array, vtk_format='binary'):
        """Register a DataArray object in p-file"""
        self._registerArrayComponent(data_array, 'PDataArray', vtk_format)


class Writer:
    """Generic XML handler for VTK files"""

    def __init__(self, vtk_format,
                 compression=None,
                 vtk_version='0.1',
                 byte_order='LittleEndian'):
        self.document = dom.getDOMImplementation()  \
                           .createDocument(None, 'VTKFile', None)
        self.root = self.document.documentElement
        self.root.setAttribute('type', vtk_format)
        self.root.setAttribute('version', vtk_version)
        self.root.setAttribute('byte_order', byte_order)
        self.data_node = self.document.createElement(vtk_format)
        self.root.appendChild(self.data_node)
        self.size_indicator_bytes = np.dtype(np.uint32).itemsize
        self.append_data_arrays = []

        if compression is not None and compression != False:
            self.root.setAttribute('compressor', 'vtkZLibDataCompressor')

            if type(compression) is not int:
                compression = -1
            else:
                if compression not in list(range(-1, 10)):
                    raise Exception(('compression level {} is not '
                                    'recognized by zlib').format(compression))
        elif not compression:
            compression = None

        self.compression = compression

    def setDataNodeAttributes(self, attributes):
        """Set attributes for the entire dataset"""
        setAttributes(self.data_node, attributes)

    def registerPiece(self, attributes={}):
        """Register a piece element"""
        return self.registerComponent('Piece', self.data_node,
                                      attributes)

    def registerComponent(self, name, parent, attributes={}):
        comp = Component(name, parent, self)
        setAttributes(comp.node, attributes)
        return comp

    def registerAppend(self):
        append_node = Component('AppendedData', self.root, self)
        setAttributes(append_node.node, {'format': 'base64'})
        self.root.appendChild(append_node.node)
        data_str = b"_"

        for data_array in self.append_data_arrays:
            data_str += encodeArray(data_array.flat_data)

        text = self.document.createTextNode(data_str.decode('ascii'))
        append_node.node.appendChild(text)

    def write(self, fd):
        if type(fd) == str:
            with open(fd, 'w') as file:
                self.write(file)
        elif issubclass(type(fd), io.TextIOBase):
            self.document.writexml(fd, indent="\n  ", addindent="  ")
        else:
            raise RuntimeError("Expected a path or "
                               + "file handle, got {}".format(type(fd)))

    def __str__(self):
        """Print XML to string"""
        return self.document.toprettyxml()
